#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

union reg {
  unsigned char *p;
  long i;
  long u;
};

#if defined(__i386__)
#  define GPREGS(_)							\
	_(a) _(b) _(c) _(d) _(si) _(di) _(bp)
#elif defined(__x86_64__)
#  define GPREGS(_)							\
	_(a) _(b) _(c) _(d) _(si) _(di) _(bp)				\
	_(r8) _(r9) _(r10) _(r11) _(r12) _(r13) _(r14) _(r15)
#elif defined(__arm__)
#  define GPREGS(_)							\
	_(r0) _(r1) _(r2) _(r3) _(r4) _(r5) _(r6) _(r7)			\
	_(r8) _(r9) _(r10) _(r11) _(r12)
#elif defined(__aarch64__)
#  define GPREGS(_)							\
	_(x0) _(x1) _(x2) _(x3) _(x4) _(x5) _(x6) _(x7)			\
	_(x8) _(x9) _(x10) _(x11) _(x12) _(x13) _(x14) _(x15)		\
	_(x16) _(x17) _(x19) _(x20) _(x21) _(x22) _(x23)		\
	_(x24) _(x25) _(x26) _(x27) _(x28) _(x29)
#else
#  error "not supported"
#endif

enum {
#define DEFCONST(r) R_##r,
  GPREGS(DEFCONST)
#undef DEFCONST
  R_flags,
  NREGS
};

static const char *const rname[] = {
#define DEFNAME(r) #r,
  GPREGS(DEFNAME)
#undef DEFNAME
  "f"
};

struct regs {
  union reg r[NREGS];
};

struct seg {
  unsigned char *p;
  size_t sz;
};

#define N(v) (sizeof(v)/sizeof((v)[0]))

#define STRCMP(a, op, b) (strcmp((a), (b)) op 0)
#define STRNCMP(a, op, b, n) (strncmp((a), (b), (n)) op 0)
#define CTYPE_HACK(func, ch) func((unsigned char)(ch))
#define ISDIGIT(ch) CTYPE_HACK(isdigit, ch)
#define ISSPACE(ch) CTYPE_HACK(isspace, ch)

#define XCHG(_)								\
  _(x00) _(x01) _(x02) _(x03) _(x04) _(x05) _(x06) _(x07)		\
  _(x08) _(x09) _(x0a) _(x0b) _(x0c) _(x0d) _(x0e) _(x0f)		\
  _(x10) _(x11) _(x12) _(x13) _(x14) _(x15) _(x16) _(x17)		\
  _(x18) _(x19) _(x1a) _(x1b) _(x1c) _(x1d) _(x1e) _(x1f)		\
  _(x20) _(x21) _(x22) _(x23) _(x24) _(x25) _(x26) _(x27)		\
  _(x28) _(x29) _(x2a) _(x2b) _(x2c) _(x2d) _(x2e) _(x2f)		\
  _(x30) _(x31) _(x32) _(x33) _(x34) _(x35) _(x36) _(x37)		\
  _(x38) _(x39) _(x3a) _(x3b) _(x3c) _(x3d) _(x3e) _(x3f)

#define DECL(x) extern const int x;
XCHG(DECL)
extern const int nop;

static const int *x[] = {
#define SLOT(x) &x,
  XCHG(SLOT)
};

extern void call_example(const int *f, struct regs *r);

static const char *prog = "???";

__attribute__((format(printf, 1, 2), noreturn))
static void barf(const char *m, ...)
{
  va_list ap;

  va_start(ap, m);
  fprintf(stderr, "%s: ", prog);
  vfprintf(stderr, m, ap);
  putc('\n', stderr);
  va_end(ap);
  exit(127);
}

static void *xmalloc(size_t sz)
{
  void *p;

  if (!sz) return (0);
  p = malloc(sz);
  if (!p) barf("malloc failed");
  return (p);
}

#define DEF_PARSEINT(name, ty, strto)					\
  static ty parse_##name(const char *what, const char *p, ty min, ty max) \
  {									\
    const char *pp = p;							\
    char *q;								\
    ty i;								\
    int err;								\
									\
    if (ISSPACE(*p)) goto bad;						\
    err = errno; errno = 0;						\
    i = strto(p, &q, 0);						\
    if (errno) goto bad;						\
    if (*q) goto bad;							\
    if (i < min || i > max) goto bad;					\
    errno = err;							\
    return (i);								\
									\
  bad:									\
    barf("bad %s `%s'", what, pp);					\
  }
DEF_PARSEINT(long, long, strtol)
DEF_PARSEINT(ulong, unsigned long, strtoul)

static int hex_digit(char ch)
{
  if ('0' <= ch && ch <= '9') return (ch - '0');
  else if ('A' <= ch && ch <= 'F') return (ch - 'A' + 10);
  else if ('a' <= ch && ch <= 'f') return (ch - 'a' + 10);
  else return (-1);
}

static void setreg(union reg *r, struct seg **seg_inout, const char *p)
{
  struct seg *seg;
  const char *pp;
  unsigned char *q;
  int hi, lo;
  size_t n;

#define LONG_REG(p) (parse_long("signed register", (p), LONG_MIN, LONG_MAX))
#define ULONG_REG(p) (parse_ulong("unsigned register", (p), 0, ULONG_MAX))

  switch (*p) {
    case '-':
      if (p[1]) r->i = LONG_REG(p);
      break;
    case 'i':
      if (p[1] != ':') goto bad;
      r->i = LONG_REG(p + 2);
      break;
    case 'u':
      if (p[1] != ':') goto bad;
      r->u = ULONG_REG(p + 2);
      break;
    case 'c':
      if (p[1] != ':' || p[3]) goto bad;
      r->u = p[2];
      break;
    case 's':
      if (p[1] != ':') goto bad;
      pp = p + 2; n = strlen(pp) + 1;
      seg = (*seg_inout)++; seg->p = xmalloc(n); seg->sz = n;
      memcpy(seg->p, pp, n); r->p = seg->p;
      break;
    case 'm':
      if (p[1] != ':') goto bad;
      pp = p + 2; n = strlen(pp); if (n%2) goto bad;
      seg = (*seg_inout)++; seg->p = q = xmalloc(n/2); seg->sz = n/2;
      while (n) {
	hi = hex_digit(pp[0]); lo = hex_digit(pp[1]);
	if (hi < 0 || lo < 0) goto bad;
	*q++ = 16*hi + lo; n -= 2; pp += 2;
      }
      r->p = seg->p;
      break;
    case 'z':
      if (p[1] != ':') goto bad;
      n = parse_ulong("buffer length", p + 2, 0, ~(size_t)0);
      seg = (*seg_inout)++; seg->p = q = xmalloc(n); seg->sz = n;
      r->p = q; memset(q, 0, n);
      break;
    default:
      if (ISDIGIT(*p)) r->u = ULONG_REG(p);
      else if (*p == '+') r->i = LONG_REG(p);
      else if (*p == '\'' && p[2] == '\'' && !p[3]) r->u = p[1];
      else goto bad;
      break;
    bad:
      barf("bad regspec `%s'", p);
  }

#undef LONG_REG
#undef ULONG_REG
}

static void dumpreg(const char *name, const union reg *r,
		    const struct seg *seg, size_t nseg)
{
  size_t i;

#if ULONG_MAX == 0xffffffff
  printf("%3s =         0x%08lx = %20ld = %20lu", name, r->u, r->i, r->u);
#else
  printf("%3s = 0x%016lx = %20ld = %20lu", name, r->u, r->i, r->u);
#endif
  if (r->u >= ' ' && r->u <= '~') printf(" = '%c'", (int)r->u);
  for (i = 0; i < nseg; i++) {
    if (r->p == seg[i].p)
      printf(" = seg[%zu] base", i);
    else if (r->p == seg[i].p + seg[i].sz)
      printf(" = seg[%zu] limit", i);
    else if (seg[i].p < r->p && r->p < seg[i].p + seg[i].sz)
      printf(" = seg[%zu] + %zu", i, (size_t)(r->p - seg[i].p));
  }
  putchar('\n');
}

static void dumpseg(const struct seg *seg)
{
  size_t i, j;
  unsigned char ch;

  for (i = 0; i < seg->sz; i += 8) {
    printf("\t%8zx :", i);
    for (j = 0; j < 8; j++)
      if (i + j >= seg->sz) printf(" **");
      else printf(" %02x", seg->p[i + j]);
    printf(" : ");
    for (j = 0; j < 8; j++)
      if (i + j >= seg->sz) putchar('*');
      else {
	ch = seg->p[i + j];
	if (' ' <= ch && ch <= '~') putchar(ch);
	else putchar('.');
      }
    putchar('\n');
  }
}

int main(int argc, char *argv[])
{
  struct regs r;
  struct seg seg[16], *segp = seg;
  size_t nseg, n;
  const char *p;
  char *q;
  unsigned long f;
  int i, j, k;
  unsigned long l;

  prog = strrchr(argv[0], '/'); if (prog) prog++; else prog = argv[0];

  if (argc < 2)
    barf("usage: %s I [REG...]",
	 prog);

  j = parse_long("program index", argv[1], -1, N(x) - 1);

#if ULONG_MAX == 0xffffffff
#  define DEAD 0xdeadbeef
#else
#  define DEAD 0xdeadbeefdeadbeef
#endif
  for (i = 0; i < NREGS - 1; i++) r.r[i].u = DEAD;
#undef DEAD
  r.r[R_flags].u = 0;

  i = 0;
  argv += 2;
  while (*argv) {
    p = *argv++;
    if (ISDIGIT(*p)) {
      l = strtoul(p, &q, 10);
      if (l < NREGS && *q == '=') { p = q + 1; i = l; }
    } else for (k = 0; k < NREGS; k++) {
      n = strlen(rname[k]);
      if (STRNCMP(p, ==, rname[k], n) && p[n] == '=')
	{ i = k; p += n + 1; break; }
    }
    if (i >= NREGS) barf("too many registers");
    setreg(&r.r[i], &segp, p); i++;
  }

  nseg = segp - seg;
  call_example(j < 0 ? &nop : x[j], &r);

  for (i = 0; i < NREGS; i++) dumpreg(rname[i], &r.r[i], seg, nseg);

  f = r.r[R_flags].u;

#if defined(__i386__) || defined(__x86_64__)

#define CF (1 <<  0)
#define PF (1 <<  2)
#define ZF (1 <<  6)
#define SF (1 <<  7)
#define OF (1 << 11)

  printf("\tstatus: %ccf %cpf %caf %czf %csf %cdf %cof\n",
	 (f >>  0)&1u ? '+' : '-',
	 (f >>  2)&1u ? '+' : '-',
	 (f >>  4)&1u ? '+' : '-',
	 (f >>  6)&1u ? '+' : '-',
	 (f >>  7)&1u ? '+' : '-',
	 (f >> 10)&1u ? '+' : '-',
	 (f >> 11)&1u ? '+' : '-');
  printf("\tcond:");
  if (f&CF) printf(" c/b/nae"); else printf(" nc/ae/nb");
  if (f&ZF) printf(" e/z"); else printf(" ne/nz");
  if (f&SF) printf(" s"); else printf(" ns");
  if (f&OF) printf(" o"); else printf(" no");
  if (f&PF) printf(" p"); else printf(" np");
  if ((f&CF) || (f&ZF)) printf(" be/na"); else printf(" a/nbe");
  if (!(f&OF) == !(f&SF)) printf(" ge/nl"); else printf(" l/nge");
  if (!(f&OF) == !(f&SF) && !(f&ZF))
    printf(" g/nle"); else printf(" le/ng");
  putchar('\n');
  printf("\tsystem: %ctf %cif iopl=%d %cnt "
			"%crf %cvm %cac %cvif %cvip %cid\n",
	 (f >>  8)&1u ? '+' : '-',
	 (f >>  9)&1u ? '+' : '-',
	 (int)((f >> 12)&1u),
	 (f >> 14)&1u ? '+' : '-',
	 (f >> 16)&1u ? '+' : '-',
	 (f >> 17)&1u ? '+' : '-',
	 (f >> 18)&1u ? '+' : '-',
	 (f >> 19)&1u ? '+' : '-',
	 (f >> 20)&1u ? '+' : '-',
	 (f >> 21)&1u ? '+' : '-');

#undef CF
#undef PF
#undef ZF
#undef SF
#undef OF

#elif defined(__arm__)

#define NF (1u << 31)
#define ZF (1u << 30)
#define CF (1u << 29)
#define VF (1u << 28)

  {
  static const char
    *modetab[] = { "?00", "?01", "?02", "?03", "?04", "?05", "?06", "?07",
		   "?08", "?09", "?10", "?11", "?12", "?13", "?14", "?15",
		   "usr", "fiq", "irq", "svc", "?20", "?21", "mon", "abt",
		   "?24", "?25", "hyp", "und", "?28", "?29", "?30", "sys" },
    *condtab[] = { "eq", "ne", "cs", "cc", "mi", "pl", "vs", "vc",
		   "hi", "ls", "ge", "lt", "gt", "le", "al", "nv" };

  printf("\tuser: %cn %cz %cc %cv %cq ge=%c%c%c%c;",
	 (f >> 31)&1u ? '+' : '-',
	 (f >> 30)&1u ? '+' : '-',
	 (f >> 29)&1u ? '+' : '-',
	 (f >> 28)&1u ? '+' : '-',
	 (f >> 27)&1u ? '+' : '-',
	 (f >> 19)&1u ? '1' : '0',
	 (f >> 18)&1u ? '1' : '0',
	 (f >> 17)&1u ? '1' : '0',
	 (f >> 16)&1u ? '1' : '0');
  if (f&NF) printf(" mi"); else printf(" pl");
  if (f&ZF) printf(" eq"); else printf(" ne");
  if (f&CF) printf(" cs/hs"); else printf(" cc/lo");
  if (f&VF) printf(" vs"); else printf(" vc");
  if ((f&CF) && !(f&ZF)) printf(" hi"); else printf(" ls");
  if (!(f&VF) == !(f&NF)) printf(" ge"); else printf(" lt");
  if (!(f&VF) == !(f&NF) && !(f&ZF)) printf(" gt"); else printf(" le");
  putchar('\n');
  printf("\tsystem: %cj it=%s:%c%c%c%c %ce %ca %ci %cf %ct m=%s\n",
	 (f >> 24)&1u ? '+' : '-',
	 condtab[(f >> 12)&15u],
	 (f >> 11)&1u ? '1' : '0',
	 (f >> 10)&1u ? '1' : '0',
	 (f >> 26)&1u ? '1' : '0',
	 (f >> 25)&1u ? '1' : '0',
	 (f >>  9)&1u ? '+' : '-',
	 (f >>  8)&1u ? '+' : '-',
	 (f >>  7)&1u ? '+' : '-',
	 (f >>  6)&1u ? '+' : '-',
	 (f >>  5)&1u ? '+' : '-',
	 modetab[(f >>  0)&31u]);
  }

#undef NF
#undef ZF
#undef CF
#undef VF

#elif defined(__aarch64__)

#define NF (1u << 31)
#define ZF (1u << 30)
#define CF (1u << 29)
#define VF (1u << 28)

  printf("\tuser: %cn %cz %cc %cv;",
	 (f >> 31)&1u ? '+' : '-',
	 (f >> 30)&1u ? '+' : '-',
	 (f >> 29)&1u ? '+' : '-',
	 (f >> 28)&1u ? '+' : '-');
  if (f&NF) printf(" mi"); else printf(" pl");
  if (f&ZF) printf(" eq"); else printf(" ne");
  if (f&CF) printf(" cs/hs"); else printf(" cc/lo");
  if (f&VF) printf(" vs"); else printf(" vc");
  if ((f&CF) && !(f&ZF)) printf(" hi"); else printf(" ls");
  if (!(f&VF) == !(f&NF)) printf(" ge"); else printf(" lt");
  if (!(f&VF) == !(f&NF) && !(f&ZF)) printf(" gt"); else printf(" le");
  putchar('\n');

#undef NF
#undef ZF
#undef CF
#undef VF

#else
#  error "not supported"
#endif

  for (i = 0; i < nseg; i++)
    { printf("seg[%d] (%p):\n", i, seg[i].p); dumpseg(&seg[i]); }

  return (0);
}
